#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <windowsx.h>
#include <string.h>
#include <iostream>
#include <fstream>

using namespace std;
char const* image_file_name;

struct bitmap_info
{
	bitmap_info() : buffer(0) {}
	~bitmap_info() { delete[] buffer; }
	unsigned long offset;
	unsigned short width;
	unsigned short height;
	COLORREF* buffer;
	static long number_of_images;
};

long bitmap_info::number_of_images = 0;


bitmap_info* LoadImageFile(char const* filename)
{
	// Open the image file, read each image creating a bitmap for each one.
	ifstream file(filename, ios::binary);
	if(!file)
	{
		return 0;
	}

	// Read image format
	long format;
	file.read((char*)&format, 4);
	bool is_565 = format == 1L;

	file.read((char*)&bitmap_info::number_of_images, 2);
	bitmap_info* bitmaps = new bitmap_info[bitmap_info::number_of_images];

	for(int i = 0; i < bitmap_info::number_of_images; ++i)
	{
		file.read((char*)&bitmaps[i].offset, 4);
		file.read((char*)&bitmaps[i].width, 2);
		file.read((char*)&bitmaps[i].height, 2);
		bitmaps[i].buffer = new COLORREF[bitmaps[i].width * bitmaps[i].height];
	}

	for(i = 0; i < bitmap_info::number_of_images; ++i)
	{
		file.seekg(bitmaps[i].offset, ios::beg);
		for(int y = 0; y < bitmaps[i].height; ++y)
		{
			for(int x = 0; x < bitmaps[i].width; ++x)
			{
				short pixel = 0;
				file.read((char*)&pixel, 2);
				const unsigned char red = is_565 ? (unsigned char)((pixel & 0xf800) >> 11) : (unsigned char)((pixel & 0x7c00) >> 10);
				const unsigned char green = is_565 ? (unsigned char)((pixel & 0x7e0) >> 5) : (unsigned char)((pixel & 0x3e0) >> 5);
				const unsigned char blue = (unsigned char)(pixel & 0x1f);
				bitmaps[i].buffer[(y * bitmaps[i].width) + x] = RGB(red * 8, green * (is_565 ? 4 : 8), blue * 8);
			}
		}
	}
	file.close();
	return bitmaps;
}

LRESULT CALLBACK WindowProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	static bitmap_info *bitmaps = 0;
	static int current_image = 0;
	switch(msg)
	{
	case WM_CREATE:
		{
			bitmaps = LoadImageFile(image_file_name);
			if(bitmaps)
			{
				// Get maximum size.
				int h = 200;
				int w = 400;
				for(int i = 0; i < bitmap_info::number_of_images; ++i)
				{
					if(bitmaps[i].height > h)
					{
						h = bitmaps[i].height;
					}
					if(bitmaps[i].width > w)
					{
						w = bitmaps[i].width;
					}
				}
				RECT crc, wrc;
				GetClientRect(hWnd, &crc);
				GetWindowRect(hWnd, &wrc);
				w += wrc.right - crc.right;
				h += wrc.bottom - crc.bottom;
				SetWindowPos(hWnd, 0, 0, 0, w, h, SWP_NOMOVE | SWP_NOZORDER);
				current_image = 0;
			}

			return 0;
		}
		break;

	case WM_KEYDOWN:
		{
			int old_image = current_image;
			if(wParam == 37)
			{
				--current_image;
				if(current_image < 0)
				{
					current_image = bitmap_info::number_of_images - 1;
				}
			}
			else if(wParam == 39)
			{
				++current_image;
				if(current_image >= bitmap_info::number_of_images)
				{
					current_image = 0;
				}
			}
			RECT rect;
			GetClientRect(hWnd, &rect);
			BOOL repaint = FALSE;
			if(bitmaps[current_image].height < bitmaps[old_image].height |
			   bitmaps[current_image].width < bitmaps[old_image].width)
			{
				repaint = TRUE;
			}
			InvalidateRect(hWnd, &rect, repaint);
			UpdateWindow(hWnd);

			return 0;
		}
		break;

	case WM_LBUTTONDOWN:

		{
			int old_image = current_image;
			++current_image;
			if(current_image >= bitmap_info::number_of_images)
			{
				current_image = 0;
			}
			RECT rect;
			GetClientRect(hWnd, &rect);
			BOOL repaint = FALSE;
			if(bitmaps[current_image].height < bitmaps[old_image].height |
			   bitmaps[current_image].width < bitmaps[old_image].width)
			{
				repaint = TRUE;
			}
			InvalidateRect(hWnd, &rect, repaint);
			UpdateWindow(hWnd);
			return 0;
		}
		break;

	case WM_RBUTTONDOWN:
		{
			int old_image = current_image;
			--current_image;
			if(current_image < 0)
			{
				current_image = bitmap_info::number_of_images - 1;
			}
			RECT rect;
			GetClientRect(hWnd, &rect);
			BOOL repaint = FALSE;
			if(bitmaps[current_image].height < bitmaps[old_image].height |
			   bitmaps[current_image].width < bitmaps[old_image].width)
			{
				repaint = TRUE;
			}
			InvalidateRect(hWnd, &rect, repaint);
			UpdateWindow(hWnd);
			return 0;
		}
		break;
	
	case WM_PAINT:
		{
			PAINTSTRUCT ps;
			HDC hdc = BeginPaint(hWnd, &ps);
			if(bitmaps)
			{
				for(int y = 0; y < bitmaps[current_image].height; ++y)
				{
					for(int x = 0; x < bitmaps[current_image].width; ++x)
					{ 
						SetPixel(hdc, x, y, bitmaps[current_image].buffer[(y * bitmaps[current_image].width) + x]);
					}
				}
			}

			EndPaint(hWnd, &ps);
			char buffer[256];
			wsprintf(buffer, "%s - %d/%d", image_file_name, current_image + 1, bitmap_info::number_of_images);
			SetWindowText(hWnd, buffer);

			return 0;
		}
		break;

	case WM_DESTROY:
		{
			PostQuitMessage(0);
			return 0;
		}
		break;

	default:
		break;
	}
	return DefWindowProc(hWnd, msg, wParam, lParam);
}

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpszCmdLine, int nCmdShow)
{
	image_file_name = lpszCmdLine;
	if(strlen(image_file_name) == 0)
	{
		::MessageBox(0, "To use this application, create a shortcut on your desktop\n"
			"and drag a Creatures 2 .s16 image file onto the shortcut.",
			"S16 Image file viewer",
			MB_OK);
		return 0;
	}

	WNDCLASS wc;
	wc.style = CS_DBLCLKS | CS_OWNDC | CS_HREDRAW | CS_VREDRAW;
	wc.lpfnWndProc = WindowProc;
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hInstance = hInstance;
	wc.hIcon = LoadIcon(0, IDI_APPLICATION);
	wc.hCursor = LoadCursor(0, IDC_ARROW);
	wc.hbrBackground = GetStockObject(BLACK_BRUSH);
	wc.lpszMenuName = 0;
	wc.lpszClassName = "S16Viewer";

	if(!RegisterClass(&wc))
	{
		return 0;
	}

	HWND hWnd = CreateWindow("S16Viewer",
		"S16 Image File Viewer",
		WS_OVERLAPPEDWINDOW | WS_VISIBLE,
		0, 0, 320, 200,
		0,
		0,
		hInstance,
		0);
	if(!hWnd)
	{
		return 0;
	}

	MSG msg;
	while(1)
	{
		if(PeekMessage(&msg, 0, 0, 0, PM_REMOVE))
		{
			if(msg.message == WM_QUIT)
			{
				break;
			}

			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}
	}

	return msg.wParam;
}


